import asyncio
import sys
import psycopg2

PY_35 = sys.version_info >= (3, 5)
PY_352 = sys.version_info >= (3, 5, 2)

if PY_35:
    from collections.abc import Coroutine

    base = Coroutine
else:
    base = object

try:
    ensure_future = asyncio.ensure_future
except AttributeError:
    ensure_future = asyncio.async


def create_future(loop):
    try:
        return loop.create_future()
    except AttributeError:
        return asyncio.Future(loop=loop)


class _ContextManager(base):
    __slots__ = ('_coro', '_obj')

    def __init__(self, coro):
        self._coro = coro
        self._obj = None

    def send(self, value):
        return self._coro.send(value)

    def throw(self, typ, val=None, tb=None):
        if val is None:
            return self._coro.throw(typ)
        elif tb is None:
            return self._coro.throw(typ, val)
        else:
            return self._coro.throw(typ, val, tb)

    def close(self):
        return self._coro.close()

    @property
    def gi_frame(self):
        return self._coro.gi_frame

    @property
    def gi_running(self):
        return self._coro.gi_running

    @property
    def gi_code(self):
        return self._coro.gi_code

    def __next__(self):
        return self.send(None)

    @asyncio.coroutine
    def __iter__(self):
        resp = yield from self._coro
        return resp

    if PY_35:
        def __await__(self):
            resp = yield from self._coro
            return resp

        @asyncio.coroutine
        def __aenter__(self):
            self._obj = yield from self._coro
            return self._obj

        @asyncio.coroutine
        def __aexit__(self, exc_type, exc, tb):
            self._obj.close()
            self._obj = None


class _SAConnectionContextManager(_ContextManager):
    if PY_35:  # pragma: no branch
        if PY_352:
            def __aiter__(self):
                return self._coro
        else:
            @asyncio.coroutine
            def __aiter__(self):
                result = yield from self._coro
                return result


class _PoolContextManager(_ContextManager):
    if PY_35:
        @asyncio.coroutine
        def __aexit__(self, exc_type, exc, tb):
            self._obj.close()
            yield from self._obj.wait_closed()
            self._obj = None


class _TransactionPointContextManager(_ContextManager):
    if PY_35:

        @asyncio.coroutine
        def __aexit__(self, exc_type, exc_val, exc_tb):
            if exc_type is not None:
                yield from self._obj.rollback_savepoint()
            else:
                yield from self._obj.release_savepoint()

            self._obj = None


class _TransactionBeginContextManager(_ContextManager):
    if PY_35:

        @asyncio.coroutine
        def __aexit__(self, exc_type, exc_val, exc_tb):
            if exc_type is not None:
                yield from self._obj.rollback()
            else:
                yield from self._obj.commit()

            self._obj = None


class _TransactionContextManager(_ContextManager):
    if PY_35:

        @asyncio.coroutine
        def __aexit__(self, exc_type, exc, tb):
            if exc_type:
                yield from self._obj.rollback()
            else:
                if self._obj.is_active:
                    yield from self._obj.commit()
            self._obj = None


class _PoolAcquireContextManager(_ContextManager):
    __slots__ = ('_coro', '_obj', '_pool')

    def __init__(self, coro, pool):
        super().__init__(coro)
        self._pool = pool

    if PY_35:
        @asyncio.coroutine
        def __aexit__(self, exc_type, exc, tb):
            yield from self._pool.release(self._obj)
            self._pool = None
            self._obj = None


class _PoolConnectionContextManager:
    """Context manager.

    This enables the following idiom for acquiring and releasing a
    connection around a block:

        with (yield from pool) as conn:
            cur = yield from conn.cursor()

    while failing loudly when accidentally using:

        with pool:
            <block>
    """

    __slots__ = ('_pool', '_conn')

    def __init__(self, pool, conn):
        self._pool = pool
        self._conn = conn

    def __enter__(self):
        assert self._conn
        return self._conn

    def __exit__(self, exc_type, exc_val, exc_tb):
        try:
            self._pool.release(self._conn)
        finally:
            self._pool = None
            self._conn = None

    if PY_35:
        @asyncio.coroutine
        def __aenter__(self):
            assert not self._conn
            self._conn = yield from self._pool.acquire()
            return self._conn

        @asyncio.coroutine
        def __aexit__(self, exc_type, exc_val, exc_tb):
            try:
                yield from self._pool.release(self._conn)
            finally:
                self._pool = None
                self._conn = None


class _PoolCursorContextManager:
    """Context manager.

    This enables the following idiom for acquiring and releasing a
    cursor around a block:

        with (yield from pool.cursor()) as cur:
            yield from cur.execute("SELECT 1")

    while failing loudly when accidentally using:

        with pool:
            <block>
    """

    __slots__ = ('_pool', '_conn', '_cur')

    def __init__(self, pool, conn, cur):
        self._pool = pool
        self._conn = conn
        self._cur = cur

    def __enter__(self):
        return self._cur

    def __exit__(self, *args):
        try:
            self._cur.close()
        except psycopg2.ProgrammingError:
            # seen instances where the cursor fails to close:
            #   https://github.com/aio-libs/aiopg/issues/364
            # We close it here so we don't return a bad connection to the pool
            self._conn.close()
            raise
        finally:
            try:
                self._pool.release(self._conn)
            finally:
                self._pool = None
                self._conn = None
                self._cur = None


if not PY_35:
    try:
        from asyncio import coroutines

        coroutines._COROUTINE_TYPES += (_ContextManager,)
    except:
        pass



import asyncio
import enum
import uuid
import warnings
from abc import ABC, abstractmethod

import psycopg2
from aiopg.utils import PY_35, _TransactionPointContextManager

__all__ = ('IsolationLevel', 'Transaction')


class IsolationCompiler(ABC):
    name = ''

    __slots__ = ('_readonly', '_deferrable')

    def __init__(self, readonly, deferrable):
        self._readonly = readonly
        self._deferrable = deferrable
        self._check_readonly_deferrable()

    def _check_readonly_deferrable(self):
        available = self._readonly or self._deferrable
        if not isinstance(self, SerializableCompiler) and available:
            raise ValueError('Is only available for serializable transactions')

    def savepoint(self, unique_id):
        return 'SAVEPOINT {}'.format(unique_id)

    def release_savepoint(self, unique_id):
        return 'RELEASE SAVEPOINT {}'.format(unique_id)

    def rollback_savepoint(self, unique_id):
        return 'ROLLBACK TO SAVEPOINT {}'.format(unique_id)

    def commit(self):
        return 'COMMIT'

    def rollback(self):
        return 'ROLLBACK'

    @abstractmethod
    def begin(self):
        raise NotImplementedError("Please Implement this method")

    def __repr__(self):
        return self.name


class ReadCommittedCompiler(IsolationCompiler):
    name = 'Read committed'

    def begin(self):
        return 'BEGIN'


class RepeatableReadCompiler(IsolationCompiler):
    name = 'Repeatable read'

    def begin(self):
        return 'BEGIN ISOLATION LEVEL REPEATABLE READ'


class SerializableCompiler(IsolationCompiler):
    name = 'Serializable'

    def begin(self):
        query = 'BEGIN ISOLATION LEVEL SERIALIZABLE'

        if self._readonly:
            query += ' READ ONLY'

        if self._deferrable:
            query += ' DEFERRABLE'

        return query


class IsolationLevel(enum.Enum):
    serializable = SerializableCompiler
    repeatable_read = RepeatableReadCompiler
    read_committed = ReadCommittedCompiler

    def __call__(self, readonly, deferrable):
        return self.value(readonly, deferrable)


class Transaction:
    __slots__ = ('_cur', '_is_begin', '_isolation', '_unique_id')

    def __init__(self, cur, isolation_level,
                 readonly=False, deferrable=False):
        self._cur = cur
        self._is_begin = False
        self._unique_id = None
        self._isolation = isolation_level(readonly, deferrable)

    @property
    def is_begin(self):
        return self._is_begin

    @asyncio.coroutine
    def begin(self):
        if self._is_begin:
            raise psycopg2.ProgrammingError(
                'You are trying to open a new transaction, use the save point')
        self._is_begin = True
        yield from self._cur.execute(self._isolation.begin())
        return self

    @asyncio.coroutine
    def commit(self):
        self._check_commit_rollback()
        yield from self._cur.execute(self._isolation.commit())
        self._is_begin = False

    @asyncio.coroutine
    def rollback(self):
        self._check_commit_rollback()
        yield from self._cur.execute(self._isolation.rollback())
        self._is_begin = False

    @asyncio.coroutine
    def rollback_savepoint(self):
        self._check_release_rollback()
        yield from self._cur.execute(
            self._isolation.rollback_savepoint(self._unique_id))
        self._unique_id = None

    @asyncio.coroutine
    def release_savepoint(self):
        self._check_release_rollback()
        yield from self._cur.execute(
            self._isolation.release_savepoint(self._unique_id))
        self._unique_id = None

    @asyncio.coroutine
    def savepoint(self):
        self._check_commit_rollback()
        if self._unique_id is not None:
            raise psycopg2.ProgrammingError('You do not shut down savepoint')

        self._unique_id = 's{}'.format(uuid.uuid1().hex)
        yield from self._cur.execute(
            self._isolation.savepoint(self._unique_id))

        return self

    def point(self):
        return _TransactionPointContextManager(self.savepoint())

    def _check_commit_rollback(self):
        if not self._is_begin:
            raise psycopg2.ProgrammingError('You are trying to commit '
                                            'the transaction does not open')

    def _check_release_rollback(self):
        self._check_commit_rollback()
        if self._unique_id is None:
            raise psycopg2.ProgrammingError('You do not start savepoint')

    def __repr__(self):
        return "<{} transaction={} id={:#x}>".format(
            self.__class__.__name__,
            self._isolation,
            id(self)
        )

    def __del__(self):
        if self._is_begin:
            warnings.warn(
                "You have not closed transaction {!r}".format(self),
                ResourceWarning)

        if self._unique_id is not None:
            warnings.warn(
                "You have not closed savepoint {!r}".format(self),
                ResourceWarning)

    if PY_35:
        @asyncio.coroutine
        def __aenter__(self):
            return (yield from self.begin())

        @asyncio.coroutine
        def __aexit__(self, exc_type, exc, tb):
            if exc_type is not None:
                yield from self.rollback()
            else:
                yield from self.commit()
